#这里是执行对CMD命令进行调用的chain
from langchain.chains.llm import LLMChain
from langchain.prompts import PromptTemplate
from lmchain.lmchain.agents import llmAgent
import os,re

class LLMCMDChain:
    def __init__(self ,llm):
        qa_prompt = PromptTemplate(template="""你现在根据需要完成对命令行的编写，要根据需求编写对应的在Windows系统终端运行的命令,不要用%question形参这种指代的参数形式,直接给出可以运行的命令。
            Question: 给我一个在Windows系统终端中可以准确执行{question}的命令。
            ,
                input_variables=["question"],
            )
            answer:""", input_variables=["question"],            )
        self.qa_chain = LLMChain(llm=llm, prompt=qa_prompt)
        self.pattern = r"```(.*?)\```"

    def run(self ,text):
        cmd_response = self.qa_chain.run(question=text)
        cmd_string = str(cmd_response).split("```")[-2][1:-1]
        os.system(cmd_string)
        return cmd_string



# LM AI Similarity Search

import warnings
warnings.filterwarnings("ignore")

import torch
import math
import asyncio
import logging
import operator
import os
import pickle
import uuid
import warnings
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sized,
    Tuple,
    Union,
)
import numpy as np
from lmchain.vectorstores.utils import DistanceStrategy, maximal_marginal_relevance

from langchain.docstore.base import AddableMixin, Docstore
from langchain.docstore.document import Document
from langchain.docstore.in_memory import InMemoryDocstore


def dependable_faiss_import(no_avx2: Optional[bool] = None) -> Any:
    """
    Import faiss if available, otherwise raise error.
    If FAISS_NO_AVX2 environment variable is set, it will be considered
    to load FAISS with no AVX2 optimization.

    Args:
        no_avx2: Load FAISS strictly with no AVX2 optimization
            so that the vectorstore is portable and compatible with other devices.
    """
    if no_avx2 is None and "FAISS_NO_AVX2" in os.environ:
        no_avx2 = bool(os.getenv("FAISS_NO_AVX2"))

    try:
        if no_avx2:
            from faiss import swigfaiss as faiss
        else:
            import faiss
    except ImportError:
        raise ImportError(
            "Could not import faiss python package. "
            "Please install it with `pip install faiss-gpu` (for CUDA supported GPU) "
            "or `pip install faiss-cpu` (depending on Python version)."
        )
    return faiss


def _len_check_if_sized(x: Any, y: Any, x_name: str, y_name: str) -> None:
    if isinstance(x, Sized) and isinstance(y, Sized) and len(x) != len(y):
        raise ValueError(
            f"{x_name} and {y_name} expected to be equal length but "
            f"len({x_name})={len(x)} and len({y_name})={len(y)}"
        )
    return



class LMASS:
    def __init__(self):
        super().__init__()
        self.texts = []
        self.vectors = []
        self.embedding_class = None

    def from_documents(self, docs, embedding_class):

        texts = [str(doc) for doc in docs]
        self.texts = texts
        self.docs = docs
        self.embedding_class = embedding_class

        self.vectors = embedding_class.embed_documents(texts)
        return self.vectors

    def from_adocuments(self, docs, embedding_class,thread_num = 5,wait_sec = 0.3):

        texts = [str(doc) for doc in docs]
        self.texts = texts
        self.docs = docs
        self.embedding_class = embedding_class

        self.vectors = embedding_class.aembed_documents(texts, thread_num=thread_num, wait_sec=wait_sec)
        return self.vectors

    def get_relevant_documents(self,query,k = 3):
        query_vector = self.embedding_class.embed_query(query)
        sorted_index = self.get_similarity_vector_indexs(query_vector,self.vectors,k = k)
        sorted_docs = [self.docs[id] for id in sorted_index]

        return sorted_docs


    def get_similarity_vector_indexs(self, query_vector ,vectors, k: int = 3,  ):
        similarity = self._cosine_similaritys(query_vector,vectors)

        #这里是按分值从大到小的进行排序
        # 使用sort()函数对tensor进行降序排序，并返回排序后的tensor和索引
        sorted_tensor, sorted_indices = torch.sort(similarity, descending=True)
        return sorted_indices[:k].numpy()


    def _cosine_similaritys(self, query_vector, vectors):
        query_vector = torch.tensor(query_vector)
        vectors = torch.tensor(vectors)
        similarity_matrix = torch.nn.functional.cosine_similarity(query_vector, vectors, dim=-1)
        return similarity_matrix

        #CosineSimilarity
    def cosine_similarity(self, query_vector, tensor_2):
        normalized_tensor_1 = query_vector / query_vector.norm(dim=-1, keepdim=True)
        normalized_tensor_2 = tensor_2 / tensor_2.norm(dim=-1, keepdim=True)
        return (normalized_tensor_1 * normalized_tensor_2).sum(dim=-1)

    #DotProductSimilarity
    def dot_product_similarity(self, query_vector, tensor_2, scale_output = True):
        result = (query_vector * tensor_2).sum(dim=-1)
        if scale_output:
            # TODO why allennlp do multiplication at here ?
            result /= math.sqrt(query_vector.size(-1))
        return result

    @classmethod
    def from_texts(
            cls,
            texts: List[str],
            embedding,
            metadatas: Optional[List[dict]] = None,
            ids: Optional[List[str]] = None,
            **kwargs: Any,
    ) :
        embeddings = embedding.embed_documents(texts)
        return cls.__from(
            texts,
            embeddings,
            embedding,
            metadatas=metadatas,
            ids=ids,
            **kwargs,
        )

    @classmethod
    def __from(
            cls,
            texts: Iterable[str],
            embeddings: List[List[float]],
            embedding,
            metadatas: Optional[Iterable[dict]] = None,
            ids: Optional[List[str]] = None,
            normalize_L2: bool = False,
            distance_strategy: DistanceStrategy = DistanceStrategy.EUCLIDEAN_DISTANCE,
            **kwargs: Any,
    ) :
        faiss = dependable_faiss_import()
        if distance_strategy == DistanceStrategy.MAX_INNER_PRODUCT:
            index = faiss.IndexFlatIP(len(embeddings[0]))
        else:
            # Default to L2, currently other metric types not initialized.
            index = faiss.IndexFlatL2(len(embeddings[0]))
        vecstore = cls(
            embedding,
            index,
            InMemoryDocstore(),
            {},
            normalize_L2=normalize_L2,
            distance_strategy=distance_strategy,
            **kwargs,
        )
        vecstore.__add(texts, embeddings, metadatas=metadatas, ids=ids)
        return vecstore



if __name__ == '__main__':
    v1 = torch.randn(size=(1,768))
    v2 = torch.randn(size=(4,768))
    lmass = LMASS()
    #print(lmass.dot_product_similarity(v1, v2))
    #print(lmass.cosine_similarity(v1, v2))
    sim_index = lmass.get_similarity_vector_indexs(v1,v2)
    print(sim_index)




import time
import logging
import requests
from typing import Optional, List, Dict, Mapping, Any

import langchain
from langchain.llms.base import LLM
from langchain.cache import InMemoryCache

logging.basicConfig(level=logging.INFO)
# 启动llm的缓存
langchain.llm_cache = InMemoryCache()


class AgentChatGLM(LLM):
    # 模型服务url
    url = "http://127.0.0.1:7866/chat"
    #url = "http://192.168.3.20:7866/chat"  #3050服务器上
    history = []

    @property
    def _llm_type(self) -> str:
        return "chatglm"

    def _construct_query(self, prompt: str) -> Dict:
        """构造请求体
        """
        query = {"query": prompt, "history": self.history}
        import json
        query = json.dumps(query)  # 对请求参数进行JSON编码

        return query

    def _construct_query_tools(self, prompt: str , tools: list ) -> Dict:
        """构造请求体
        """
        tools_info = {"role": "system",
                      "content": "你现在是一个查找使用何种工具以及传递何种参数的工具助手，你会一步步的思考问题。你根据需求查找工具函数箱中最合适的工具函数，然后返回工具函数名称和所工具函数对应的参数，参数必须要和需求中的目标对应。",
                      "tools": tools}
        query = {"query": prompt, "history": tools_info}
        import json
        query = json.dumps(query)  # 对请求参数进行JSON编码

        return query


    @classmethod
    def _post(self, url: str,              query: Dict) -> Any:

        """POST请求"""
        response = requests.post(url, data=query).json()
        return response

    def _call(self, prompt: str,              stop: Optional[List[str]] = None, tools:list = None) -> str:
        """_call"""
        if tools == None:
            # construct query
            query = self._construct_query(prompt=prompt)

            # post
            response = self._post(url=self.url,query=query)

            response_chat = response["response"];
            self.history = response["history"]

            return response_chat
        else:

            query = self._construct_query_tools(prompt=prompt,tools=tools)
            # post
            response = self._post(url=self.url, query=query)
            self.history = response["history"]  #这个history要放上面
            response = response["response"]
            try:
                #import ast
                #response = ast.literal_eval(response)
                ret = tool_register.dispatch_tool(response["name"], response["parameters"])
                response_chat = llm(prompt=ret)
            except:
                response_chat = response
            return str(response_chat)

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters.
        """
        _param_dict = {
            "url": self.url
        }
        return _param_dict


if __name__ == "__main__":

    import tool_register

    # 获取注册后的全部工具，并以json的形式返回
    tools = tool_register.get_tools()
    "--------------------------------------首先是对tools的定义---------------------------------------"

    llm = AgentChatGLM()
    llm.url = "http://192.168.3.20:7866/chat"
    while True:
        while True:
            human_input = input("Human: ")
            if human_input == "tools":
                break

            begin_time = time.time() * 1000
            # 请求模型
            response = llm(human_input)
            end_time = time.time() * 1000
            used_time = round(end_time - begin_time, 3)
            #logging.info(f"chatGLM process time: {used_time}ms")
            print(f"Chat: {response}")

        human_input = input("Human_with_tools_Ask: ")
        response = llm(prompt=human_input,tools=tools)
        print(f"Chat_with_tools_Que: {response}")




